/***********************************************************************
 adriamassanet@gmail.com
 TFC UOC 2013
 NoFeud Social CryptoPrototype
 *************************************************************************/

function RSAKey_serialize (onlyPublicKey)
{
    if (this.d==null || onlyPublicKey)
    {
        return {
            n: this.n.toString(16),
            e: this.e.toString(16)
        };
    }
    else
    {
        return {
            n: this.n.toString(16),
            e: this.e.toString(16),
            d: this.d.toString(16),
            p: this.p.toString(16),
            q: this.q.toString(16),
            dmp1: this.dmp1.toString(16),
            dmq1: this.dmq1.toString(16),
            coeff: this.coeff.toString(16)
        };
    }
};

function RSAKey_unserialize(rsaKeyEnc)
{
    if ( "d" in rsaKeyEnc)
    {
        this.setPrivateEx(rsaKeyEnc.n, rsaKeyEnc.e, rsaKeyEnc.d, rsaKeyEnc.p, rsaKeyEnc.q, rsaKeyEnc.dmp1, rsaKeyEnc.dmq1, rsaKeyEnc.coeff);
    } else
    {
        this.setPublic(rsaKeyEnc.n, rsaKeyEnc.e);
    }
}

function RSAKey_isPrivate()
{
    return  ( "d" in rsaKeyEnc);
}

function RSAKey_signSHA1(data)
{
    var sHashHex = hex_sha1(data); // 40 chars
    var biMessage = new BigInteger(sHashHex,16);
    var biSign = this.doPrivate(biMessage);
    return biSign.toString(16);
}

function RSAKey_verifySHA1(data,signature)
{
    var sHashHex = hex_sha1(data); // 40 chars
    var biSignature = new BigInteger(signature,16);
    var biOriginal = this.doPublic(biSignature);
    var original = biOriginal.toString(16);
    return sHashHex==original;
}

function RSAKey_envelope(plaintext)
{
    var randomKey =  secrets.random(32);
    var cipheredKey=this.encrypt(randomKey);
    var cipheredData=CryptoJS.AES.encrypt(plaintext, randomKey).toString();
    return { cipheredKey: cipheredKey , cipheredData: cipheredData };
}
function RSAKey_unenvelope(ciphertext)
{
    var randomKey=this.decrypt(ciphertext.cipheredKey);
    return CryptoJS.AES.decrypt(ciphertext.cipheredData, randomKey).toString(CryptoJS.enc.Utf8);
}


RSAKey.prototype.isPrivate = RSAKey_isPrivate
RSAKey.prototype.serialize = RSAKey_serialize;
RSAKey.prototype.unserialize = RSAKey_unserialize;
RSAKey.prototype.signSHA1 = RSAKey_signSHA1;
RSAKey.prototype.verifySHA1 = RSAKey_verifySHA1;
RSAKey.prototype.envelope = RSAKey_envelope;
RSAKey.prototype.unenvelope = RSAKey_unenvelope;

function Certificate ( )
{
    this.getToBeSigned = function()
    {
        return {
            issuer : this.issuerKey.serialize(true),
            subject : this.subject,
            idProof : this.idProof,
            publicKey : this.key.serialize(true)
        };
    }

    this.generateSelfSigned = function (subject, idProof)
    {
        this.key = new RSAKey();
        this.key.generate(512,'010001');
        this.issuerKey = this.key;
        this.subject    = subject;
        this.idProof    = idProof;
        this.signature  = this.key.signSHA1(this.getToBeSigned());
    }

    this.issueCertificate = function (subject)
    {
        issuedCertificate = new Certificate();
        issuedCertificate.key = new RSAKey();
        issuedCertificate.key.generate(512,'010001');
        issuedCertificate.issuerKey  = this.key;
        issuedCertificate.subject    = subject;
        issuedCertificate.idProof    = "parent-signed";
        issuedCertificate.signature  = this.key.signSHA1(issuedCertificate.getToBeSigned());
        return issuedCertificate;
    }
    this.verify = function (issuedCert)
    {
        if (!this.key.verifySHA1(issuedCert.getToBeSigned(),issuedCert.signature))
        {
            alert("Failed to verify certfificate signature of "+issuedCert.subject);
            return false;
        }
        return true;
    }

    this.serialize = function ( )
    {
        var serialized =
        {
            certId    : this.id(),
            issuerPublicKey : this.issuerKey.serialize(true),
            subject   : this.subject,
            publicKey : this.key.serialize(true),
            idProof   : this.idProof,
            signature : this.signature
        };
        return serialized;
    }

    this.unserialize = function (serialized)
    {
        this.subject = serialized.subject;
        this.signature = serialized.signature;
        this.idProof = serialized.idProof;
        this.key = new RSAKey();
        this.key.unserialize(serialized.publicKey);
        this.issuerKey = new RSAKey();
        this.issuerKey.unserialize(serialized.issuerPublicKey);

        if (!this.issuerKey.verifySHA1(this.getToBeSigned(),this.signature))
        {
            alert("Cannot verify SHA1 signature of certificate "+this.subject);
            return false;
        }
        return true;
    }
    this.encrypt = function (data)
    {
        return this.key.encrypt(data);
    }

    this.decrypt = function (data)
    {
        return this.key.decrypt(data);
    }

    this.id = function ()
    {
        return hex_sha1(this.key.n.toString(16));
    }
    this.issuerId = function ()
    {
        return hex_sha1(this.issuerKey.n.toString(16));
    }

    this.idProof          = null;
    this.issuerKey        = null;
    this.subject          = null;
    this.signature        = null;
    this.key              = null;
}

function SCryptoV2 ( )
{
    this.setMyKeyCertificate = function(cert)
    {
        this.cert = cert;
        this.fiendsKeyCertificates[cert.id()] =cert;
    }
    this.addFriendMasterPbkCertificate = function(cert)
    {
        this.fiendsMasterCertificates[cert.id()]=  cert ;
    }
    this.addFriendKeyPvkCertificate = function (keyCertificate)
    {
        if (keyCertificate.issuerId() in this.fiendsMasterCertificates)
        {
            var masterCertificate = this.fiendsMasterCertificates[keyCertificate.issuerId()];
            masterCertificate.verify(keyCertificate);
            this.fiendsKeyCertificates[keyCertificate.id()] = keyCertificate;
        } else alert("failed addFriendKeyCertificate (master certificate not found)");
    }
    this.setMinShare = function (minShare)
    {
        this.minShare = minShare;
    }
    this.createNewSecretAndGetSharesVectorInfo = function() {
        this.currentSecret = secrets.random(32);

        var publishVector=Array();
        var shares = secrets.share(this.currentSecret, Object.keys(this.fiendsKeyCertificates).length, this.minShare);
        shareNo = 0;
        for (var keyId in this.fiendsKeyCertificates)
        {
            if (keyId!=cert.id())
            {
                var randomKey =  secrets.random(32);
                var cipheredKey=this.fiendsKeyCertificates[keyId].key.encrypt(randomKey);
                var cipheredShare=CryptoJS.AES.encrypt(shares[shareNo].toString(), randomKey).toString();
                publishVector.push( { id: keyId, cipheredKey: cipheredKey , cipheredShare: cipheredShare });
                shareNo++;
            }
        }
        return { publisherpbk : this.cert.id() , shares: publishVector };
    };


    this.cipherMessage = function (plainText)
    {
        return CryptoJS.AES.encrypt(plainText, this.currentSecret).toString();
    };

    this.decipherMessage = function (publishVectorInfo, cipherText)
    {
        var publishVector = publishVectorInfo.shares;
        var foundShares = [];
        for (shareNo =0; shareNo<publishVector.length;++shareNo)
        {
            var shareId = publishVector[shareNo].id;
            for (var keyId in this.fiendsKeyCertificates)
            {
                if(keyId==shareId)
                {
                    var key=this.fiendsKeyCertificates[keyId].decrypt(publishVector[shareNo].cipheredKey);
                    var share = CryptoJS.AES.decrypt(publishVector[shareNo].cipheredShare, key);
                    foundShares.push(share.toString(CryptoJS.enc.Utf8));
                    break;
                }
            }
        }
        var secret = secrets.combine(foundShares);
        return CryptoJS.AES.decrypt(cipherText, secret).toString(CryptoJS.enc.Utf8);
    };
    this.selftest = function ()
    {
        A = new Certificate();
        A.generateSelfSigned("A","");
        A1 = A.issueCertificate("A1");

        B = new Certificate();
        B.generateSelfSigned("B","");
        B1 = B.issueCertificate("B1");

        C = new Certificate();
        C.generateSelfSigned("C","");
        C1 = C.issueCertificate("C1");

        D = new Certificate();
        D.generateSelfSigned("D","");
        D1 = D.issueCertificate("D1");

        Z = new Certificate();
        Z.generateSelfSigned("Z","");
        Z1 = Z.issueCertificate("Z1");

        ACrypto = new SCryptoV2();
        ACrypto.setMyKeyCertificate(A1);
        ACrypto.addFriendMasterPbkCertificate(B);
        ACrypto.addFriendMasterPbkCertificate(C);
        ACrypto.addFriendMasterPbkCertificate(D);
        ACrypto.addFriendKeyPvkCertificate(B1);
        ACrypto.addFriendKeyPvkCertificate(C1);
        ACrypto.addFriendKeyPvkCertificate(D1);

        ZCrypto = new SCryptoV2();
        ZCrypto.setMyKeyCertificate(Z1);
        ZCrypto.addFriendMasterPbkCertificate(B);
        ZCrypto.addFriendMasterPbkCertificate(C);
        ZCrypto.addFriendMasterPbkCertificate(D);
        ZCrypto.addFriendKeyPvkCertificate(B1);
        ZCrypto.addFriendKeyPvkCertificate(C1);
        ZCrypto.addFriendKeyPvkCertificate(D1);

        m = "hola";
        var A_sharesVector = ACrypto.createNewSecretAndGetSharesVectorInfo();
        var A_cipheredMessage = ACrypto.cipherMessage(m);
        var deciphered = ZCrypto.decipherMessage(A_sharesVector,A_cipheredMessage);
        return deciphered == m;
    }

    this.constructor = function ()
    {
    }

    this.cert = null;
    this.fiendsMasterCertificates = [];
    this.fiendsKeyCertificates = [];
    this.minShare = 2;
    this.currentSecret = null;
    this.init = this.constructor()
};

